/*
bgs_demo -i ../../dataset/video.tar -a 100 -o ../../output -c ../../config/centerConfigFile.xml -p ../../config/camParam.xml

TODO
> input = tar of images!

#### Background Subtraction Methods #####

 Codebook is new!
 Tapter is TBoy!

 List of all algorithms:
 AdaptiveBackgroundLearning,AdaptiveSelectiveBackgroundLearning,CodeBook,DPAdaptiveMedian,DPEigenbackground,
 DPGrimsonGMM,DPMean,DPPratiMediod,DPTexture,DPWrenGA,DPZivkovicAGMM,FrameDifference,FuzzyChoquetIntegral,
 FuzzySugenoIntegral,GMG,IndependentMultimodal,KDE,KNN,LBAdaptiveSOM,LBFuzzyAdaptiveSOM,LBFuzzyGaussian,
 LBMixtureOfGaussians,LBP_MRF,LBSimpleGaussian,LOBSTER,MixtureOfGaussianV2,MixtureOfGaussianV1,MultiCue,
 MultiLayer,PAWCS,PixelBasedAdaptiveSegmenter,SigmaDelta,StaticFrameDifference,SuBSENSE,T2FGMM_UM,T2FGMM_UV,
 T2FMRF_UM,T2FMRF_UV,TwoPoints,ViBe,VuMeter,WeightedMovingMean,WeightedMovingVariance

 (Note that some of these algorithms are available only for a specific version of OpenCV, see algorithms.h)

see paper https://dl.acm.org/citation.cfm?id=2321600
https://ieeexplore.ieee.org/document/4527178/
was in benchmark on top https://www.researchgate.net/publication/259340906_A_comprehensive_review_of_background_subtraction_algorithms_evaluated_with_synthetic_and_real_videos
my own adapter to use the model (tapter?)
*/

//cpp c
#include <iostream>
#include <algorithm>
#include <cstdlib>
#include <stdio.h>      /* printf, scanf, puts, NULL */
#include <stdlib.h>     /* srand, rand */
#include <time.h>       /* time */
#include <ctime>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <archive.h>
#include <archive_entry.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>

#include <regex> //cpp11
#include <opencv2/opencv.hpp> //opencv

#include "tools.h" //tools
#include "algorithms.h" //bgs

#define PROCESS_CENTER_VERSION_MAJOR 2
#define PROCESS_CENTER_VERSION_MINOR 0
#define PROCESS_CENTER_VERSION_MINOR_FIXES 1


using namespace bgslibrary::algorithms;
using namespace cv;
using namespace std;


const string ABOUT_STRING=R"("Camtron Software B v2.0.0

the software extracts the position of individuals in a single frame
work-flow:
a) load image
b) make lens correction for image with calibration data
c) uses background subtraction method “PixelBasedAdaptiveSegmenter” to get region of interest in picture Hofmann et. a. 2012  “Background Segmentation with Feedback: The Pixel-Based Adaptive Segmenter”
(framework bgslib BGSLibrary version 2.0.0 )
(choosen because TOP 5 in paper Andrews Sobral 2014 et. al.  “A comprehensive review of background subtraction algorithms evaluated with synthetic and real videos”)
d) apply gaussian filter to smooth the image  (framework opencv version 3.2.)
d) use canny edge detection to extract polygon of interest
Canny 1986. A Computational Approach to Edge Detection
(framework opencv version 3.2.)
e) made a selection of all extracted polygons to reduce wrong positives
remove all polygons with area of bounding rectangles lower threshold min ( 150 pixel x pixel)
and bigger than threshold max ( 300 000 pixel x pixel  )
f) calculate convex hull of all remaining points of all polygones
g) calculate Hu invariants of convex hull
Hu 1962 “Visual Pattern Recognition by Moment Invariants “
(framework opencv version 3.2.)
g) calculate the Centroid of convex hull with hu invariants
e) save the remaining centroids for a frame to file
)";

//
// FWD DCLs
//

void my_handler(int signum);
vector<char> copyDataInBuffer(struct archive *aw);
int toFrameNumber(string filename); //! convert our actual jpg file number to framenumber
void read_camParamFile();
void read_centerFile();
void libArchive_init();


//
// GLOBALS
//

String camParamFile, centerFile, outputDir, inputFile;
int amountFiles;

Mat distCoeffs;
Mat cameraMatrix;

int circleCenterX=880;
int circleCenterY=750;
int circleRadius=700;

struct archive *archive;
struct archive *ext;
struct archive_entry *entry;

int g_badSignalFlagAbort = 0;
Scalar color_red( 0,0,255,255 );
bool testMode;

namespace error {
	enum {
		argument_parse_error,
		unexpected_frame_number,
		file_not_found,
		config_filedir,
		archive_file,
		archive_read_header,
		archive_write,

	};
}


// ###########

void parse(int argc, char** argv)
{
	const String keys =
	"{help h usage ? || " + String(argv[0]) + " -i=in.tar -a=10 -c=center.xml -o=outPath -p=camparam.xml }"
	"{i input|../input/| input tarfile/dir containing tars? }"
	"{o output|../output| output path / dir }"
	"{a amount|| amount of frames }"
	"{p param|../config/camparam.yml| camera parameter file }"
	"{c center|../config/center.xml| exact center xml-conf file }"
	"{t testMode|false|test mode for debugging purposes. eg. only process 10 frames}"
	;

	CommandLineParser parser(argc, argv, keys);
	parser.about(ABOUT_STRING);

	if (parser.has("help"))
	{
		parser.printMessage();
		exit(0);
	}

	inputFile = parser.get<String>("i");
	outputDir = parser.get<String>("o");
	amountFiles = parser.get<int>("a");
	camParamFile = parser.get<String>("p");
	centerFile = parser.get<String>("c");

	if (parser.has("testMode"))
		testMode = parser.get<bool>("t");

	if (!parser.check())
	{
		parser.printErrors();
		exit(error::argument_parse_error);
	}

	cout << "amountFiles: " << amountFiles << endl;
	cout << boolalpha << "testMode: " << testMode << endl;

	if( !Tools::check({
		"camParamFile", camParamFile,
		"centerFile", centerFile,
		"inputFile", inputFile,
		"outputDir", outputDir
	}))
		exit(error::config_filedir);

	Tools::maybeAppendSepToDir(outputDir);
	Tools::print_ruler();

	exit(-1);
}

void read_camParamFile()
{
	FileStorage fs(camParamFile, FileStorage::READ);
	if (!fs.isOpened())
	{
		std::cerr << "error opening '" << camParamFile << "' will abort" << endl;
		exit(error::config_filedir);
	}

	fs["camera_matrix"] >>  cameraMatrix;
	fs["distortion_coefficients"] >> distCoeffs;

	cout <<	"cameraMatrix" << cameraMatrix << endl;
	cout << "distCoeffs" << distCoeffs << endl;
	Tools::print_ruler();
} //fs.release();

void read_centerFile()
{
	 //! read the center config file to cut out the ROI
	// TODO merge this config with the Tapter.xml ??
	FileStorage fs(centerFile, FileStorage::READ);
	if (!fs.isOpened())
	{
		cout << "error opening :" << centerFile << endl;
		exit(error::config_filedir);
	}

	circleCenterX = (int) fs["circleCenterX"];
	circleCenterY = (int) fs["circleCenterY"];
	circleRadius = (int) fs["circleRadius"];

	cout  << "circleRadius: " << circleRadius << endl;
	cout  << "circleCenterX: " << circleCenterX << endl;
	cout  << "circleCenterY: " << circleCenterY << endl;
	Tools::print_ruler();

} //fs.release();

void libArchive_init()
{
	int flags = ARCHIVE_EXTRACT_TIME; // see https://linux.die.net/man/3/archive_write_disk for more flags
	const char *filename = inputFile.c_str(); //XXX needed?
	// "testFile" || "data_sized.tar"

	archive = archive_read_new();
	ext = archive_write_disk_new();
	archive_write_disk_set_options(ext, flags);

	archive_read_support_format_tar(archive);

	// get the filename
	if (filename != NULL && strcmp(filename, "-") == 0)
		filename = NULL;

	// and open the handler
	if (archive_read_open_filename(archive, filename, 10240))
	{
		cerr << "archive_read_open_filename: error: " << archive_error_string(archive) << " will abort" << endl;
		exit(error::archive_file);
	}
}

int main(int argc, char * argv[])
{
	cout << "Using OpenCV " << CV_MAJOR_VERSION << "." << CV_MINOR_VERSION << "." << CV_SUBMINOR_VERSION << endl;
	cout << "Using processcenter  " << PROCESS_CENTER_VERSION_MAJOR << "." << PROCESS_CENTER_VERSION_MINOR << "." << PROCESS_CENTER_VERSION_MINOR_FIXES << endl << endl;

	signal(SIGUSR1, my_handler);
	srand (time(NULL)); //init the random number generator

	parse(argc, argv);

	read_camParamFile();
	read_centerFile();
	libArchive_init();

	// IBGS *bgs = new FrameDifference;
	IBGS *bgs = new PixelBasedAdaptiveSegmenter; // TODO


	cout << "a) process all frames " << endl;

	int everyPic = 60*5;
	int frameCounter = 0;
	int myFileTarWriterHeaderCounter = 0;
	int frameCounterOld = -1;
	int r = -1;
	vector<string> myFileErrorList;//list for pure io error
	vector<string> myFileListNoContour; //list for no contours found;
	vector<string> myFileListAfterContourSelection; //list for no contours found after selection;

	clock_t beginAll = clock();

	// loop over _all_ pictures in the tar archive
	for (;;)
	{

		//measure time consumption
		//sa opencv tick timer:
		//	https://docs.opencv.org/4.x/d9/d6f/classcv_1_1TickMeter.html#details
		clock_t begin = clock();  //for every single file
		vector<char> vec;

		//read the next header
		r = archive_read_next_header(archive, &entry);
		if (r == ARCHIVE_EOF)
			break;
		else if (r != ARCHIVE_OK)
		{
			cerr << "archive_read_next_header: error: " << archive_error_string(archive) << " will abort" << endl;
			exit(error::archive_read_header);
		}
		cout << "here2" << endl;

		// get filename
		const char *fileNamePtr = archive_entry_pathname(entry);
		string filename(fileNamePtr,strlen(fileNamePtr) );
		cout << "fileName: " << filename << endl;
		// convert it to a framenumber
		frameCounter = toFrameNumber(filename);
		cout << "frameCounter: " << frameCounter << endl;

		if(frameCounterOld>frameCounter)
		{
			cerr << " error during read in the file number, found non-monotonic order, will abort";
			exit(error::unexpected_frame_number);
		}

		frameCounterOld = frameCounter;

		r = archive_write_header(ext, entry);
		if (r != ARCHIVE_OK)
		{
			cerr << "archive_write_header() error: " << archive_error_string(ext)<< endl;
			myFileTarWriterHeaderCounter++;
			continue; // overjump all in for loop
		}

		vec = copyDataInBuffer(archive);
		if(vec.empty())
			cerr << "error during load data into buffer";

		r = archive_write_finish_entry(ext);
		if (r != ARCHIVE_OK)
		{
			cerr << "archive_write_finish_entry: error: " << archive_error_string(ext) << " will abort" << endl;
			exit(error::archive_write);
		}

		// read the image buffer as a jpg picture and decode it
		Mat img_input = imdecode(Mat(vec), 1);

		cout << "\t" << frameCounter << "\tof \t" << amountFiles << "  load file :" << filename << endl;

		//of data is present
		if(img_input.data)
		{

			//! step 0 correct lense distortion
			Mat imgLenseCorrection = img_input.clone();
			//see https://docs.opencv.org/2.4/modules/imgproc/doc/geometric_transformations.html#undistort
			cv::undistort(img_input, imgLenseCorrection, cameraMatrix, distCoeffs);

			//! step 1) cut out a smaller ROI
			img_input = Tools::cropImageCircle(imgLenseCorrection,circleCenterX,circleCenterY,circleRadius);

			//! step 2) normal bgs processing
			cv::Mat img_mask;
			cv::Mat img_bkgmodel;

			bgs->process(img_input, img_mask, img_bkgmodel); // by default, it shows automatically the foreground mask image

			//! step 3) apply edge detection
			//TODO make this in a function, how many times is this listed ??
			//TODO read from tapter config
			//TODO all parameter from applyCannyEdgeAndCalcCountours also !!
			double threshholdMin = 150;
			double threshholdMax = 200;
			int apertureSize = 3;
			vector<vector<Point> > contours = Tools::applyCannyEdgeAndCalcCountours(img_mask,threshholdMin,threshholdMax,apertureSize);

			//define what will be written
			vector<vector<Point> > contourSelection;
			vector<Point2f> massCenters;
			vector<Point> conHull;
			Point2f muConvexHullMassCenter(0.0,0.0);

			if(!contours.empty())
			{

				//! step 4) we make a selection out of all contours with area sizes
				//******************************

				//calc all min rotated  rectangles for all contour from canny edge detect
				//exclude very small/big
				vector<RotatedRect> minRect( contours.size() );

				//calc boxes around the contours
				for( size_t i = 0; i < contours.size(); i++ )
					minRect[i] = minAreaRect( Mat(contours[i]) ); //may use boundingRect ?? to use fix non rotated rectangles ?

				vector<Point2f> recCenterPoints;
				float areaMinThreshold = 150;
				float areaMaxThreshold = 300000; //old max threshold was to small //TODO apply moving filter ??, an more adaptive approach

				//iterate all rectangles
				for( size_t i = 0; i< minRect.size(); i++ )
				{
					Point2f rect_points[4];
					//get all points of the retange
					minRect[i].points( rect_points );

					//construct contour based on rectangle points because contour != rectangle with points
					vector<Point> contourRect;
					for(int j=0;j<4;j++)
						contourRect.push_back(rect_points[j]);

					//calc the area of the contour
					double area0 = contourArea(contourRect);
					//over stepp all small areas
					if(area0<areaMinThreshold||area0>areaMaxThreshold)
					{
						//cout << i << ":area0:" << area0 << " dismissed " << endl;
						//skipe if the area is too small
						continue;
					}
					else
					{
						//cout << i << ":area0:" << area0 << " choose " << endl;

						//get the center of this rectangle
						Point2f center = minRect[i].center;
						recCenterPoints.push_back(center);

						//we also add the this contour to a selection
						contourSelection.push_back(contours[i]);

					}
				}//end iterate all min rectangle

				if(!contourSelection.empty())
				{
					//! step 5) we calc the moments from the contour selection
					vector<Moments> mu(contourSelection.size() );
					for( size_t i = 0; i < contourSelection.size(); i++ )
					{
						mu[i] = moments( contourSelection[i], false );
					}

					//  Get the mass centers:
					massCenters =  vector<Point2f>( contourSelection.size() );
					for( size_t i = 0; i < contourSelection.size(); i++ )
					{
						massCenters[i] = Point2f( mu[i].m10/mu[i].m00 , mu[i].m01/mu[i].m00 );
					}

					//! step 6) calc convex hull of all points of the contour selection *********
					//TODO produce center of convex hull of all polygones
					//TODO double check if ne contour sharing a center point ? nearby ??

					//merge all points
					vector<Point> allContourPoints;
					for (size_t cC = 0; cC < contourSelection.size(); ++cC)
						for(size_t cP =0; cP < contourSelection[cC].size(); cP++)
						{
							Point currentContourPixel = contourSelection[cC][cP];
							allContourPoints.push_back(currentContourPixel);
						}

					// calc the hull ******************
						conHull = vector<Point>(allContourPoints.size());
						convexHull( Mat(allContourPoints), conHull, false );

					//calc the min circle around
					//    Point roiCenter;
					//    float roiRadius;
					//minEnclosingCircle(conHull,roiCenter,roiRadius);

					// calc the mass center of the convex hull. see:
					// https://docs.opencv.org/4.x/d8/d23/classcv_1_1Moments.html#details
						Moments muConvexHull;
						if(!conHull.empty())
						{
							muConvexHull = moments(conHull, true );
							muConvexHullMassCenter = Point2f( muConvexHull.m10/muConvexHull.m00 , muConvexHull.m01/muConvexHull.m00 );
						}
					}
				else //end after selection is empty
				{
					cerr << "error, no contour found after selection in file: " << filename << endl;
					myFileListAfterContourSelection.push_back(filename);
				}
			}//end if contours are empty, to canny edge found nothing
			else
			{
				cerr << "error, no contour found at all in file: " << filename << endl;
				myFileListNoContour.push_back(filename);

			}

#ifdef  MC_SHOW_STEP_ANALYSE

			if(!conHull.empty())
			{
				Mat imgConvexHull = Mat::zeros( img_input.size(), CV_8UC3 );
				imgConvexHull =  Scalar(255,255,255); //fill the picture

				polylines(imgConvexHull, hullComplete, true, color_red, 1, 8);

				//draw circle around
				//circle( imgConvexHull, roiCenter, (int) roiRadius, color_red, 2, 8, 0 );

				// draw it
				convert2 << outputDir << Tools::mNzero(frameCounter) << "_convex_hull.jpg";
				cv::imwrite(convert2.str().c_str(), imgConvexHull);
			}
			else
				cout << "convex hull has no points will skip file: " << i << endl;

#endif

			//cout << " found center at : " << massCenters.at(0).x << ";" << massCenters.at(0).y << endl;
			//! step 8: we write down all our results in yml file
			string nameOutPutFileData =  outputDir + Tools::mNzero(frameCounter) + ".yml";

			//cout << "output file:" << nameOutPutFileData;

			FileStorage fs(nameOutPutFileData.c_str(), FileStorage::WRITE);
			fs << "masscenters" << massCenters;
			fs << "polygonselection" << contourSelection;
			fs << "convexhull" << conHull;
			fs << "masscenterconvexhull" << muConvexHullMassCenter;
			fs.release();

			//! occasionaly write a dbg picture
			if(frameCounter%everyPic==0)
			{
				RNG rng(4344234);
				Mat imgDebugPaint2 = Mat::zeros( img_input.size(), CV_8UC3 );
				imgDebugPaint2 =  Scalar(255,255,255); //fill the picture white

				// write all polyies of the selection and the mass centers with a random color
				for( size_t i = 0; i< contourSelection.size(); i++ )
				{
					//random color
					Scalar color = Scalar( rng.uniform(0, 255), rng.uniform(0,255), rng.uniform(0,255) );
					//contour
					drawContours( imgDebugPaint2, contourSelection, i, color, 1, LINE_AA);
					//draw the center
					circle( imgDebugPaint2, massCenters[i], 4, color, -1, 8, 0 );
				}

				// write the convex hull
				if(!conHull.empty())
				{
					//the poly
					polylines(imgDebugPaint2, conHull, true, color_red, 1, 8);
					//the center
					circle( imgDebugPaint2,muConvexHullMassCenter, 4, color_red, -1, 8, 0 );
				}

				// copy
				Mat imgOverlay2 = img_input.clone();
				// add overlay of paintings
				addWeighted( imgDebugPaint2, 0.7, imgOverlay2, 0.3, 0.0, imgOverlay2);
				// writeout file
				string nameOutPutFileDBGpic =  outputDir + Tools::mNzero(frameCounter) + string(".jpg");

				imwrite(nameOutPutFileDBGpic.c_str(),imgOverlay2);

			}//end if dbg picture


		}//end if the loaded picture has data
		else
		{
			cerr << "error loading file: " << filename << ", will skip this file" << endl;
			myFileErrorList.push_back(filename);
		}

		//calc the time which we used for a picture //TODO move out of loop
		clock_t end = clock();
		double elapsedSecs = double(end - begin) / CLOCKS_PER_SEC;


		//calc the time which was used for all pictures
		clock_t endAll = clock();
		double elapsedSecTotal = double(endAll - beginAll) / CLOCKS_PER_SEC;
		cout << "process single pic:\t" << elapsedSecs << " s  - \t\t" << (int)(elapsedSecTotal/60)<< " min -\t" << (int)(elapsedSecTotal/60/60)<< " h" << endl;

		if(g_badSignalFlagAbort)
			frameCounter = amountFiles; // abort

		if(testMode && frameCounter>10)
			exit(0);

	}//end for ever loop

	//finishing time
	clock_t endAll = clock();
	double elapsed = double(endAll - beginAll) / CLOCKS_PER_SEC;
	cout << "process : " << amountFiles << " files took:\t" << (int)(elapsed/60)<< " min -\t" << (int)(elapsed/60/60)<< " h \n in total" << endl;

	//    //we write the random file list to a file
	//    string nameOutRandomFile =  outputDir + "randlist.yml";
	//    FileStorage fs(nameOutRandomFile.c_str(), FileStorage::WRITE);
	//    fs << "randomlist" << myRandomTrainList;
	//    fs.release();

	//TODO we should merge the file

	cout << "amount of libarchive read header errors :" << myFileTarWriterHeaderCounter;

	// write the random file list to a file
	string nameOutErrorList =  outputDir + "fileErrorList.yml";
	cout << "amount of file errors: " << myFileErrorList.size()<< endl;
	FileStorage fs2(nameOutErrorList.c_str(), FileStorage::WRITE);
	fs2 << "fileErrorIOs" << myFileErrorList;

	// write error list no contours found in file
	cout << "amount contour errors with canny edge: " << myFileListNoContour.size() << endl;
	fs2 << "fileErrorNoContours" << myFileListNoContour;

	// write error list no contours found in file
	cout << "amount contour errors after selection: " << myFileListAfterContourSelection.size() << endl;
	fs2 << "fileErrorNoContoursAfterSelections" << myFileListAfterContourSelection;

	// bgs related things
	delete bgs;

	// opencv related things
	fs2.release();

	cvDestroyAllWindows();

	//close lib archive related things
	archive_read_close(archive);
	archive_read_free(archive);

	archive_write_close(ext);
	archive_write_free(ext);

	return 0;
}



vector<char> copyDataInBuffer(struct archive *aw)
{
	int r;
	const void *buff;
	vector<char> vec;
	size_t size;
#if ARCHIVE_VERSION_NUMBER >= 3000000
	int64_t offset;
#else
	off_t offset;
#endif
	//forever till we get a specific return
	for (;;) {
		r = archive_read_data_block(aw, &buff, &size, &offset);
		if (r == ARCHIVE_EOF)
		{
			return vec; // everything fine, were are just at the end
		}

		if (r != ARCHIVE_OK)
		{
			cerr << "error during read archive " << endl;
			return vec;
		}

		//a good discussion :  https://stackoverflow.com/questions/259297/how-do-you-copy-the-contents-of-an-array-to-a-stdvector-in-c-without-looping
		// Method 4: vector::insert
		{
			//we rename our pointer to avoid a weird compiler warning
			char *foo = (char*) buff;
			vec.insert(vec.end(), &foo[0], &foo[size]);
		}

	}

	return vec;
}

void my_handler(int signum)
{
	if (signum == SIGUSR1)
	{
		g_badSignalFlagAbort = 1;
		cerr << "receive signal to abort" << endl;
	}
}

int toFrameNumber(string filename)
{
	int retVal = -1;

	//cut out all non-needed substrings
	filename = regex_replace(filename, regex("/"), "");
	filename = regex_replace(filename, regex("data_sized"), "");
	filename = regex_replace(filename, regex("jpg"), "");
	filename = regex_replace(filename, regex("\\."), "");

	retVal = stoi( filename );

	//cout << "filename out:" << filename << " number: " << retVal << endl;

	return retVal;
}
